from dataclasses import asdict, is_dataclass
from datetime import date, datetime
from typing import Callable, Dict, T, Type, Union

from flask import Response, json


class Serializer():
    '''Provides serialization for Flask Hintful.

    Default serializers:
        dict: flask.json.dumps,
        str: str,
        int: str,
        float: str,
        bool: str,
        date: lambda d: d.isoformat(),
        datetime: lambda d: d.isoformat(),

    Dataclasses and classes with a __marshmallow__ attribute are also supported.
    '''

    def __init__(self):
        self.serializers: Dict[Type, Callable] = {
            dict: lambda d: json.dumps(d, default=isodate_json_encoder),
            str: str,
            int: str,
            float: str,
            bool: str,
            date: lambda d: d.isoformat(),
            datetime: lambda d: d.isoformat(),
        }

    def add_serializer(self, type_: Type, serializer_func: Callable):
        '''Adds a serializer for type `type_`

        `serializer_func` must be able to serialize any instance of `type_` into a str.

        Args:
            type_ (Type): Any type
            serializer_func (Callable): Callable that can serialize a str/dict into type `t`
        '''
        self.serializers[type_] = serializer_func

    def serialize_response(self, data: T) -> Union[str, tuple, Response]:
        '''Serializes `data` into a response Flask understands.
        If Content-Type was supplied pass the same ahead to Flask, otherwise
        uses 'application/json' as the default Content-Type

        Args:
            data (T): data to be serialized, a tuple return like Flask`s or a Flask Response object.

        Returns:
            Union[str, tuple, Response]: Serialized response in a way Flask understands
        '''
        if isinstance(data, tuple):
            headers: Dict[str, str] = {}
            status = None

            if len(data) == 3:
                body, status, headers = data
            elif len(data) == 2:
                if isinstance(data[1], (int, str)):
                    body, status = data
                else:
                    body, headers = data
            if headers is None or headers.get('Content-Type') is None:
                headers['Content-Type'] = 'application/json'

            if status is not None:
                return self.serialize(body), status, headers
            return self.serialize(body), headers

        if isinstance(data, Response):
            return data
        return self.serialize(data), {'Content-Type': 'application/json'}

    def serialize(self, data: T) -> str:
        '''Serializes `data` into a string using the registered serializers that matches data type.
        Uses `is_dataclass` to determine if `data` is a dataclass, if positive uses `serialize_dataclass`.
        Uses `is_marshmallow_model` to determine if `data` is a model, if positive
        uses `serialize_marshmallow_model`

        Args:
            data (Any): Data to be serialized as a string

        Raises:
            TypeError: If there arent any registered serializers for data
            and data is not a dataclass nor marshmallow model.

        Returns:
            str: string representation of data
        '''
        serializer = self.serializers.get(data.__class__)
        if serializer is not None:
            return serializer(data)
        if self.is_list(data):
            return self.serialize_list(data)
        if self.is_dataclass(data):
            return self.serialize_dataclass(data)
        if self.is_marshmallow_model(data):
            return self.serialize_marshmallow_model(data)
        raise TypeError(f'Cannot serialize type {data.__class__}')

    @staticmethod
    def is_dataclass(data: T) -> bool:
        '''Determines if data is a dataclass or not using dataclasses.is_dataclass

        Args:
            data (T): Any python object

        Returns:
            bool: True if data is a dataclass, False otherwise
        '''
        if is_dataclass(data):
            return True
        return False

    def serialize_dataclass(self, data: T) -> str:
        '''Uses dataclasses.asdict to transform `data` into a dict, then uses the
        registered dict serializer to serialize this dataclass

        Args:
            data (T): A python dataclass

        Returns:
            str: string representation of data
        '''
        return self.serializers.get(dict, json.dumps)(asdict(data))

    @staticmethod
    def serialize_dataclass_to_dict(data: T) -> dict:
        return asdict(data)

    @staticmethod
    def is_marshmallow_model(data: T) -> bool:
        '''Determines if data is a marshmallow object by checking if it has a marshmallow
        schema in __marshmallow__ attribute

        Args:
            data (T): Any python object

        Returns:
            bool: True if data is a marshmallow model, False otherwise
        '''
        if hasattr(data, '__marshmallow__'):
            return True
        return False

    @staticmethod
    def serialize_marshmallow_model(data: T) -> str:
        '''Uses marshmallow.Schema.dumps to serialize `data`. Assumes that
        data makes it's schema available in __marshmallow__ attrbute

        Args:
            data (T): A marshmallow model

        Returns:
            str: string representation of data
        '''
        return data.__marshmallow__().dumps(data).data

    @staticmethod
    def serialize_marshmallow_model_to_dict(data: T) -> dict:
        '''Uses marshmallow.Schema.dumps to serialize `data`. Assumes that
        data makes it's schema available in __marshmallow__ attrbute

        Args:
            data (T): A marshmallow model

        Returns:
            str: string representation of data
        '''
        return data.__marshmallow__().dump(data).data

    @staticmethod
    def is_list(data: T) -> bool:
        '''Determines if data is a list or not

        Args:
            data (T): Any python object

        Returns:
            bool: True if data is a list , False otherwise
        '''
        if isinstance(data, list):
            return True
        return False

    def serialize_list(self, data: T) -> str:
        '''Serializes list and as items of this list

        Args:
            data (T): A python list

        Returns:
            str: Serialized list with serialized items
        '''
        for i, item in enumerate(data):
            if self.is_dataclass(item):
                data[i] = self.serialize_dataclass_to_dict(item)
            elif self.is_marshmallow_model(item):
                data[i] = self.serialize_marshmallow_model_to_dict(item)
        return json.dumps(data, default=isodate_json_encoder)


def isodate_json_encoder(data):
    if isinstance(data, (date, datetime)):
        return data.isoformat()



from inspect import getdoc, signature
from typing import Callable


def get_func_sig(func: Callable) -> dict:
    sig = signature(func, follow_wrapped=True)
    return {
        "return": sig.return_annotation,
        "params": sig.parameters,
        "doc": getdoc(func),
        "empty": sig.empty
    }



from functools import wraps
from typing import Callable

from flask import request

from .deserializer import Deserializer
from .serializer import Serializer
from .utils import get_func_sig


def view_func_wrapper(view_func: Callable, serializer: Serializer, deserializer: Deserializer):
    '''Wraps around the view_func to deserialize Flask request view args, args and
    body as parameters for the view_func and serialize the view_func return.

    Args:
        view_func (Callable): Function that will be wrapped
        serializer (Serializer): Serializer to serialize response
        deserializer (Deserializer): Deserializer to deserialize args
    '''
    func_sig = get_func_sig(view_func)

    @wraps(view_func)
    def decorator(**_):
        args = request.args.copy()
        args.update(request.view_args)
        deserialized_args = deserializer.deserialize_args(
            args, func_sig['params'], request.get_json()
        )
        response = view_func(**deserialized_args)
        return serializer.serialize_response(response)
    return decorator


class BlueprintWrapper():
    '''This class is used to help wrap view funcs registered using a Flask Blueprint.

    Args:
        app (FlaskHintful): FlaskHintful api to register the Blueprints routes on
        url_prefix (str, optional): [description]. Defaults to ''.
    '''

    def __init__(self, app, url_prefix: str = ''):
        self.app = app
        self.url_prefix = url_prefix

    def add_url_rule(self, rule, endpoint, view_func, **options):
        '''Wraps view_func with view_func_wrapper, then return a lambda expression as is
        expected by Flask Blueprint`s deferred_functions.
        '''
        wrapped_view_func = view_func_wrapper(view_func, self.app.serializer, self.app.deserializer)
        if self.url_prefix:
            prefixed_rule = '/'.join((self.url_prefix.rstrip('/'), rule.lstrip('/')))
        self.app.openapi_provider.add_openapi_path(
            rule or prefixed_rule, options.get('methods', ['GET']), view_func)
        return lambda s: s.add_url_rule(rule, endpoint, wrapped_view_func, **options)



import re
from dataclasses import is_dataclass
from typing import Callable, List

from flask import Response, current_app, jsonify
from openapi_specgen import OpenApi, OpenApiParam, OpenApiPath, OpenApiResponse

from .utils import get_func_sig


class OpenApiProvider():
    '''Provides automatically generation of OpenApi specification for registered paths.
    '''

    def __init__(self):
        self.openapi_paths: List[OpenApiPath] = []

    def add_openapi_path(self, rule: str, methods: List[str], view_func: Callable):
        '''Add a new OpenApi Path for each method. Inspects view_func's type hints to be able to determine
        the appropriate paramater types.

            Args:
                rule (str): HTTP Path that the view_func will be registered in Flask
                methods (List[str]): List of HTTP Methods this path can receive
                view_func (Callable): Function that is called when this HTTP path is invoked
        '''
        func_sig = get_func_sig(view_func)
        openapi_params = []
        body = None

        for param_name, param in func_sig['params'].items():
            if hasattr(param.annotation, '__marshmallow__'):
                body = param.annotation.__marshmallow__
            elif is_dataclass(param.annotation):
                body = param.annotation
            elif f'<{param_name}>' in re.findall('<.*?>', rule):
                openapi_params.append(
                    OpenApiParam(
                        param_name,
                        'path',
                        data_type=param.annotation if param.annotation is not param.empty else str,
                        default=param.default if param.default is not param.empty else None,
                        required=param.default is param.empty
                    )
                )
            else:
                openapi_params.append(
                    OpenApiParam(
                        param_name,
                        'query',
                        data_type=param.annotation if param.annotation is not param.empty else str,
                        default=param.default if param.default is not param.empty else None,
                        required=param.default is param.empty
                    )
                )

        response_type = func_sig['return'] if func_sig['return'] is not func_sig['empty'] else str

        if hasattr(response_type, '__marshmallow__'):
            response_type = response_type.__marshmallow__

        openapi_response = OpenApiResponse('', data_type=response_type)

        for method in methods:
            self.openapi_paths.append(
                OpenApiPath(
                    rule.replace('<', '{').replace('>', '}'),
                    method.lower(),
                    [openapi_response],
                    openapi_params,
                    descr=func_sig['doc'],
                    request_body=body
                )
            )

    def get_openapi_spec(self) -> Response:
        '''Generates the OpenApi specification based on all registered Paths.

        Returns:
            Response: A Flask response containing the OpenApi spec as json
        '''
        return jsonify(OpenApi(current_app.name, self.openapi_paths).as_dict())

    @staticmethod
    def get_openapi_ui() -> str:
        '''OpenApi UI provided by SwaggerUI

        Returns:
            str: SwaggerUI html
        '''
        openapi_json_path = current_app.config.get('FLASK_HINTFUL_OPENAPI_JSON_URL', '/openapi.json')
        return '''
                <! doctype html>
                <html>
                <head>
                <link type="text/css" rel="stylesheet" href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui.css">
                <title>
                </title>
                </head>
                <body>
                <div id="swagger-ui">
                </div>
                <script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui-bundle.js"></script>
                <!-- `SwaggerUIBundle` is now available on the page -->
                <script>

                const ui = SwaggerUIBundle({{
                    url: '{}',
                    dom_id: '#swagger-ui',
                    presets: [
                        SwaggerUIBundle.presets.apis,
                        SwaggerUIBundle.SwaggerUIStandalonePreset
                    ],
                    layout: "BaseLayout",
                    deepLinking: true
                }})
                </script>
                </body>
                </html>
            '''.format(openapi_json_path)




from dataclasses import fields, is_dataclass
from datetime import date, datetime
from typing import Callable, Dict, List, T, Type, Union

from dateutil.parser import parse as date_parser
from flask import json


class Deserializer():
    '''Provides deserialization for Flask Hintful.

    Default deserializers:
        dict: flask.json.loads,
        str: str,
        int: int,
        float: float,
        bool: flask_hintful.deserializer.str_to_bool,
        datetime: dateutil.parser.parse,
        date: lambda d: date_parser(d).date()

    Dataclasses and classes with a __marshmallow__ attribute are also supported.
    '''

    def __init__(self):
        self.deserializers: Dict[Type, Callable] = {
            dict: json.loads,
            str: str,
            int: int,
            float: float,
            bool: str_to_bool,
            datetime: date_parser,
            date: lambda d: date_parser(d).date()
        }

    def add_deserializer(self, type_: Type, deserializer_func: Callable):
        '''Adds a deserializer for type `type_`

        `serializer_func` must be able to deserialize any str or dict and return an
        instance of type `type_`

        Args:
            type_ (Type): Any python type
            deserializer_func (Callable):
        '''
        self.deserializers[type_] = deserializer_func

    def deserialize_args(self, args, params, body=None) -> dict:
        '''Deserializes all args and body by finding the expected type's from params.
        Args that are found in params are ignored, unless params contains a VAR_KEYWORD
        param (e.g, **kwargs).
        Since body does not have a name assumes that any param that is a Dataclass or Marshmallow Model
        is supposed to receive the body.

        Args:
            args ([werkzeug.datastructures.MultiDict]): Args from a Flask request
            body ([str]): JSON Body from a Flask request
            params ([dict]): Parameters from inspect.signature

        Returns:
            dict: A dict with all deserialized args
        '''
        deserialized_args = {}
        for param_name, param in params.items():
            if param_name in args:
                arg = args.poplist(param_name)
                deserialized_args[param_name] = arg if len(arg) > 1 else arg[0]
                if param.annotation != param.empty:
                    deserialized_args[param_name] = self.deserialize(arg, param.annotation)
            else:
                if param.kind == param.VAR_KEYWORD:
                    deserialized_args.update(args)
                if body is not None:
                    if self.is_dataclass(param.annotation) or self.is_marshmallow_model(param.annotation):
                        deserialized_args[param_name] = self.deserialize(
                            body, params[param_name].annotation
                        )
        return deserialized_args

    def deserialize(self, data: Union[List, str, dict], type_: Type[T]) -> T:
        '''Deserializes `data` into an instance of `type_` using the registered
        deserializer that matches the type of `type_`.
        If data is a dataclass recursively serializes all fields and passes a dict to the default constructor.
        If data has an attribute __marshmallow__ assumes it's a Marshmallow Schema and uses Schema.load()

        Raises:
            TypeError: If there arent any registered deserializers for data
            and data is not a dataclass nor has a __marshmallow__ attribute.

        Args:
            data (Union[List, str, dict]): Data to be deserialized as type_
            type_ (T): Any type

        Returns:
            T: An instance of type_
        '''
        deserializer = self.deserializers.get(type_)
        if issubclass(type_, list):
            return data
        if isinstance(data, list):
            data = data[0]
        if deserializer is not None:
            return deserializer(data)
        if self.is_dataclass(type_):
            return self.deserialize_dataclass(data, type_)
        if self.is_marshmallow_model(type_):
            return self.deserialize_marshmallow_model(data, type_)
        raise TypeError(f'Cannot deserialize type {type_}')

    @staticmethod
    def is_dataclass(type_: Type) -> bool:
        '''Determines if type_ is a dataclass or not using dataclasses.is_dataclass

        Args:
            data (T): Any python object

        Returns:
            bool: True if data is a dataclass, False otherwise
        '''
        if is_dataclass(type_):
            return True
        return False

    def deserialize_dataclass(self, data: Union[str, dict], type_: Type[T]) -> T:
        '''Uses dataclasses.asdict to transform `data` into a dict, then uses the
        registered dict serializer to serialize this dataclass

        Args:
            data (T): A python dataclass

        Returns:
            str: string representation of data
        '''
        if isinstance(data, str):
            parsed_data = json.loads(data)
        else:
            parsed_data = data
        for field in fields(type_):
            if parsed_data.get(field.name) and field.type not in [str, int, float, bool]:
                parsed_data[field.name] = self.deserialize(parsed_data[field.name], field.type)
        return type_(**parsed_data)

    @staticmethod
    def is_marshmallow_model(data: T) -> bool:
        '''Determines if data is a marshmallow object by checking if it has a marshmallow
        schema in __marshmallow__ attribute

        Args:
            data (T): Any python object

        Returns:
            bool: True if data is a marshmallow model, False otherwise
        '''
        if hasattr(data, '__marshmallow__'):
            return True
        return False

    @staticmethod
    def deserialize_marshmallow_model(data: Union[str, dict], type_: Type) -> str:
        '''Uses marshmallow.Schema.load or loads to deserialize `data`. Assumes that
        data makes it's schema available in __marshmallow__ attrbute

        Args:
            data (T): A marshmallow model

        Returns:
            str: string representation of data
        '''
        if isinstance(data, str):
            return type_.__marshmallow__().loads(data).data
        return type_.__marshmallow__().load(data).data


def str_to_bool(data: str) -> bool:
    '''Parse data into bool.

    Args:
        data (str): Any string

    Raises:
        ValueError: If str is not in either TRUE_STRS or FALSE_STRS

    Returns:
        bool: True if str is in TRUE_STRS, False if str is in FALSE_STRS.
    '''
    if data.lower() in TRUE_STRS:
        return True
    if data.lower() in FALSE_STRS:
        return False
    raise ValueError(f'{data} not in accepted values {TRUE_STRS}, {FALSE_STRS}')


TRUE_STRS: List[str] = ['true', '1', 't', 'y']
FALSE_STRS: List[str] = ['false', '0', 'f', 'n']



import pathlib

from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="flask-hintful",
    version="0.0.3dev6",
    description="Flask extension for generating restful apis using type hints to automatically (de)serialize parameters and generate openapi docs.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/GabrielCappelli/flask-hintful",
    author="Gabriel Cappelli",
    author_email="6148081+GabrielCappelli@users.noreply.github.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["flask_hintful"],
    include_package_data=True,
    install_requires=['flask', 'openapi-specgen'],
    entry_points={}
)



from flask import Blueprint, Flask

from .deserializer import Deserializer
from .openapi import OpenApiProvider
from .serializer import Serializer
from .wrapper import BlueprintWrapper, view_func_wrapper


class FlaskHintful():
    '''The FlaskHintful object implements `route` and `register_blueprints` that mimic those of Flask. These
    will wrap your view funcs to serialize/deserialize HTTP query/path/body and pass them as params to your
    view functions.

    It inspect types hints in your view funcs to attempt to serialize/deserialize arguments to the expected
    types.

    It will also inspect all registered routes and automatically generate a OpenApi specification.

    Args:
        flask_app (Flask): Instance of the underlying Flask application
        serializer (Serializer, optional): Serialization provider. Defaults to Serializer().
        deserializer (Deserializer, optional): Deserialization provider. Defaults to Deserializer().
        openapi_provider (OpenApiProvider, optional): OpenApiProvider. Defaults to OpenApiProvider().
    '''

    def __init__(self,
                 flask_app: Flask,
                 serializer=None,
                 deserializer=None,
                 openapi_provider=None
                 ):
        self.flask_app = flask_app
        self.serializer = serializer or Serializer()
        self.deserializer = deserializer or Deserializer()
        self.openapi_provider = openapi_provider or OpenApiProvider()
        self.flask_app.add_url_rule(
            flask_app.config.get('FLASK_HINTFUL_OPENAPI_JSON_URL', '/openapi.json'),
            view_func=self.openapi_provider.get_openapi_spec
        )
        self.flask_app.add_url_rule(
            flask_app.config.get('FLASK_HINTFUL_OPENAPI_UI_URL', '/swagger'),
            view_func=self.openapi_provider.get_openapi_ui
        )

    def route(self, rule: str, **options):
        '''Wrap the decorated function using view_func_wrapper then register the wrapped func
        within the underlying Flask application.

        Args:
            rule (str): HTTP path to register this view func.
        '''
        def decorator(view_func):
            wrapped_view_func = view_func_wrapper(
                view_func,
                self.serializer,
                self.deserializer
            )
            self.flask_app.route(rule, **options)(wrapped_view_func)
            self.openapi_provider.add_openapi_path(rule, options.get('methods', ['GET']), view_func)
            return view_func
        return decorator

    def register_blueprint(self, blueprint: Blueprint):
        '''Wraps all view funcs declared on blueprint using BlueprintWrapper, then registers the
        Blueprint within the underlying Flask application.

        Args:
            blueprint (Blueprint): Flask Blueprint
        '''
        bp_wrapper = BlueprintWrapper(self, blueprint.url_prefix)
        for i, func in enumerate(blueprint.deferred_functions):
            blueprint.deferred_functions[i] = func(bp_wrapper)
        self.flask_app.register_blueprint(blueprint)
